# Copyright (c) 2019-2025 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""U2F security key handler"""

from base64 import urlsafe_b64encode
import ctypes
from hashlib import sha256
import hmac
import time
from typing import Callable, List, Mapping, NoReturn, Optional
from typing import Sequence, Tuple, TypeVar, cast


_PollResult = TypeVar('_PollResult')
_SKResidentKey = Tuple[int, str, bytes, bytes]


_CTAP1_POLL_INTERVAL = 0.1

_dummy_hash = 32 * b'\0'

# Flags
SSH_SK_USER_PRESENCE_REQD = 0x01

# Algorithms
SSH_SK_ECDSA = -7
SSH_SK_ED25519 = -8


def _decode_public_key(alg: int, public_key: Mapping[int, object]) -> bytes:
    """Decode algorithm and public value from a CTAP public key"""

    result = cast(bytes, public_key[-2])

    if alg == SSH_SK_ED25519:
        return  result
    else:
        return  b'\x04' + result + cast(bytes, public_key[-3])


def _verify_rp_id(_rp_id: str, _origin: str):
    """Allow any relying party name -- SSH encodes the application here"""

    return True


def _ctap1_poll(poll_interval: float, func: Callable[..., _PollResult],
                *args: object) -> _PollResult:
    """Poll until a CTAP1 response is received"""

    while True:
        try:
            return func(*args)
        except ApduError as exc:
            if exc.code != APDU.USE_NOT_SATISFIED:
                raise

            time.sleep(poll_interval)


def _ctap1_enroll(dev: 'CtapHidDevice', alg: int,
                  application: str) -> Tuple[bytes, bytes]:
    """Enroll a new security key using CTAP version 1"""

    ctap1 = Ctap1(dev)

    if alg != SSH_SK_ECDSA:
        raise ValueError('Unsupported algorithm')

    app_hash = sha256(application.encode('utf-8')).digest()
    registration = _ctap1_poll(_CTAP1_POLL_INTERVAL, ctap1.register,
                               _dummy_hash, app_hash)

    return registration.public_key, registration.key_handle


def _ctap2_enroll(dev: 'CtapHidDevice', alg: int, application: str,
                  user: str, pin: Optional[str],
                  resident: bool) -> Tuple[bytes, bytes]:
    """Enroll a new security key using CTAP version 2"""

    ctap2 = Ctap2(dev)

    rp = {'id': application, 'name': application}
    user_cred = {'id': user.encode('utf-8'), 'name': user}
    key_params = [{'type': 'public-key', 'alg': alg}]
    options = {'rk': resident}

    pin_protocol: Optional[PinProtocolV1]
    pin_auth: Optional[bytes]

    if pin:
        pin_protocol = PinProtocolV1()
        pin_token = ClientPin(ctap2, pin_protocol).get_pin_token(pin)
        pin_auth = hmac.new(pin_token, _dummy_hash, sha256).digest()[:16]
    else:
        pin_protocol = None
        pin_auth = None

    pin_version = pin_protocol.VERSION if pin_protocol else None
    cred = ctap2.make_credential(_dummy_hash, rp, user_cred, key_params,
                                 options=options, pin_uv_param=pin_auth,
                                 pin_uv_protocol=pin_version)
    cdata = cred.auth_data.credential_data

    # pylint: disable=no-member
    return _decode_public_key(alg, cdata.public_key), cdata.credential_id


def _win_enroll(alg: int, application: str, user: str) -> Tuple[bytes, bytes]:
    """Enroll a new security key using Windows WebAuthn API"""

    data_collector = DefaultClientDataCollector(origin=application,
                                                verify=_verify_rp_id)
    client = WindowsClient(data_collector)

    rp = {'id': application, 'name': application}
    user_cred = {'id': user.encode('utf-8'), 'name': user}
    key_params = [{'type': 'public-key', 'alg': alg}]
    options = {'rp': rp, 'user': user_cred, 'challenge': b'',
               'pubKeyCredParams': key_params}

    result = client.make_credential(options)
    response = result.response
    cdata = response.attestation_object.auth_data.credential_data

    # pylint: disable=no-member
    return _decode_public_key(alg, cdata.public_key), cdata.credential_id


def _ctap1_sign(dev: 'CtapHidDevice', message_hash: bytes, application: str,
                key_handle: bytes) -> Tuple[int, int, bytes]:
    """Sign a message with a security key using CTAP version 1"""

    ctap1 = Ctap1(dev)

    app_hash = sha256(application.encode('utf-8')).digest()

    auth_response = _ctap1_poll(_CTAP1_POLL_INTERVAL, ctap1.authenticate,
                                message_hash, app_hash, key_handle)

    flags = auth_response[0]
    counter = int.from_bytes(auth_response[1:5], 'big')
    sig = auth_response[5:]

    return flags, counter, sig


def _ctap2_sign(dev: 'CtapHidDevice', message_hash: bytes,
                application: str, key_handle: bytes,
                touch_required: bool) -> Tuple[int, int, bytes]:
    """Sign a message with a security key using CTAP version 2"""

    ctap2 = Ctap2(dev)

    allow_creds = [{'type': 'public-key', 'id': key_handle}]
    options = {'up': touch_required}

    # See if key handle exists before requiring touch
    if touch_required:
        ctap2.get_assertions(application, message_hash, allow_creds,
                             options={'up': False})

    assertion = ctap2.get_assertions(application, message_hash, allow_creds,
                                     options=options)[0]

    auth_data = assertion.auth_data

    return auth_data.flags, auth_data.counter, assertion.signature


def _win_sign(data: bytes, application: str,
              key_handle: bytes) -> Tuple[int, int, bytes, bytes]:
    """Sign a message with a security key using Windows WebAuthn API"""

    data_collector = DefaultClientDataCollector(origin=application,
                                                verify=_verify_rp_id)
    client = WindowsClient(data_collector)

    creds = [{'type': 'public-key', 'id': key_handle}]
    options = {'challenge': data, 'rpId': application,
               'allowCredentials': creds}

    result = client.get_assertion(options).get_response(0)
    response = result.response
    auth_data = response.authenticator_data

    return auth_data.flags, auth_data.counter, \
           response.signature, bytes(response.client_data)


def sk_webauthn_prefix(data: bytes, application: str) -> bytes:
    """Calculate a WebAuthn request prefix"""

    return b'{"type":"webauthn.get","challenge":"' + \
           urlsafe_b64encode(data).rstrip(b'=') + b'","origin":"' + \
           application.encode('utf-8') + b'"'


def sk_enroll(alg: int, application: str, user: str, pin: Optional[str],
              resident: bool) -> Tuple[bytes, bytes]:
    """Enroll a new security key"""

    if sk_use_webauthn:
        return _win_enroll(alg, application, user)

    try:
        dev = next(CtapHidDevice.list_devices())
    except StopIteration:
        raise ValueError('No security key found') from None

    try:
        return _ctap2_enroll(dev, alg, application, user, pin, resident)
    except CtapError as exc:
        if exc.code == CtapError.ERR.PUAT_REQUIRED:
            raise ValueError('PIN required') from None
        elif exc.code == CtapError.ERR.PIN_INVALID:
            raise ValueError('Invalid PIN') from None
        else:
            raise ValueError(str(exc)) from None
    except ValueError:
        try:
            return _ctap1_enroll(dev, alg, application)
        except ApduError as exc:
            raise ValueError(str(exc)) from None
    finally:
        dev.close()


def sk_sign(data: bytes, application: str, key_handle: bytes, flags: int,
            is_webauthn: bool = False) -> Tuple[int, int, bytes, bytes]:
    """Sign a message with a security key"""

    touch_required = bool(flags & SSH_SK_USER_PRESENCE_REQD)

    if is_webauthn and sk_use_webauthn:
        return _win_sign(data, application, key_handle)

    if is_webauthn:
        data = sk_webauthn_prefix(data, application) + b'}'

    message_hash = sha256(data).digest()

    for dev in CtapHidDevice.list_devices():
        try:
            flags, counter, sig = _ctap2_sign(dev, message_hash, application,
                                              key_handle, touch_required)

            return flags, counter, sig, data
        except CtapError as exc:
            if exc.code != CtapError.ERR.NO_CREDENTIALS:
                raise ValueError(str(exc)) from None
        except ValueError:
            try:
                flags, counter, sig = _ctap1_sign(dev, message_hash,
                                                  application, key_handle)

                return flags, counter, sig, data
            except ApduError as exc:
                if exc.code != APDU.WRONG_DATA:
                    raise ValueError(str(exc)) from None
        finally:
            dev.close()

    raise ValueError('Security key credential not found')


def sk_get_resident(application: str, user: Optional[str],
                    pin: str) -> Sequence[_SKResidentKey]:
    """Get keys resident on a security key"""

    app_hash = sha256(application.encode('utf-8')).digest()
    result: List[_SKResidentKey] = []

    for dev in CtapHidDevice.list_devices():
        try:
            ctap2 = Ctap2(dev)

            pin_protocol = PinProtocolV1()
            pin_token = ClientPin(ctap2, pin_protocol).get_pin_token(pin)
            cred_mgmt = CredentialManagement(ctap2, pin_protocol, pin_token)

            for cred in cred_mgmt.enumerate_creds(app_hash):
                user_info = cast(Mapping[str, object],
                                 cred[CredentialManagement.RESULT.USER])
                name = cast(str, user_info['name'])

                if user and name != user:
                    continue

                cred_id = cast(Mapping[str, object],
                               cred[CredentialManagement.RESULT.CREDENTIAL_ID])
                key_handle = cast(bytes, cred_id['id'])

                public_key = cast(Mapping[int, object],
                                  cred[CredentialManagement.RESULT.PUBLIC_KEY])

                alg = cast(int, public_key[3])
                public_value = _decode_public_key(alg, public_key)

                result.append((alg, name, public_value, key_handle))
        except CtapError as exc:
            if exc.code == CtapError.ERR.NO_CREDENTIALS:
                continue
            elif exc.code == CtapError.ERR.PIN_INVALID:
                raise ValueError('Invalid PIN') from None
            elif exc.code == CtapError.ERR.PIN_NOT_SET:
                raise ValueError('PIN not set') from None
            else:
                raise ValueError(str(exc)) from None
        finally:
            dev.close()

    return result


try:
    from fido2.client import DefaultClientDataCollector
    from fido2.ctap import CtapError
    from fido2.ctap1 import Ctap1, APDU, ApduError
    from fido2.ctap2 import Ctap2, ClientPin, PinProtocolV1
    from fido2.ctap2 import CredentialManagement
    from fido2.hid import CtapHidDevice

    sk_available = True
except (ImportError, OSError, AttributeError): # pragma: no cover
    sk_available = False

    def _sk_not_available(*args: object, **kwargs: object) -> NoReturn:
        """Report that security key support is unavailable"""

        raise ValueError('Security key support not available')

    sk_enroll = _sk_not_available
    sk_sign = _sk_not_available
    sk_get_resident = _sk_not_available

try: # pragma: no cover
    from fido2.client.windows import WindowsClient

    sk_use_webauthn = WindowsClient.is_available() and \
                      hasattr(ctypes, 'windll') and \
                      not ctypes.windll.shell32.IsUserAnAdmin()
except (AttributeError, ImportError):
    WindowsClient = None # type: ignore
    sk_use_webauthn = False



# Copyright (c) 2019-2023 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""A shim around PyCA for Edwards-curve keys and key exchange"""

from typing import Dict, Optional, Union, cast

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends.openssl import backend
from cryptography.hazmat.primitives.asymmetric import ed25519, ed448
from cryptography.hazmat.primitives.asymmetric import x25519, x448
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import PrivateFormat
from cryptography.hazmat.primitives.serialization import PublicFormat
from cryptography.hazmat.primitives.serialization import NoEncryption

from .misc import CryptoKey, PyCAKey


_EdPrivateKey = Union[ed25519.Ed25519PrivateKey, ed448.Ed448PrivateKey]
_EdPublicKey = Union[ed25519.Ed25519PublicKey, ed448.Ed448PublicKey]


ed25519_available = backend.ed25519_supported()
ed448_available = backend.ed448_supported()
curve25519_available = backend.x25519_supported()
curve448_available = backend.x448_supported()


class _EdDSAKey(CryptoKey):
    """Base class for shim around PyCA for EdDSA keys"""

    def __init__(self, pyca_key: PyCAKey, pub: bytes,
                 priv: Optional[bytes] = None):
        super().__init__(pyca_key)

        self._pub = pub
        self._priv = priv

    @property
    def public_value(self) -> bytes:
        """Return the public value encoded as a byte string"""

        return self._pub

    @property
    def private_value(self) -> Optional[bytes]:
        """Return the private value encoded as a byte string"""

        return self._priv


class EdDSAPrivateKey(_EdDSAKey):
    """A shim around PyCA for EdDSA private keys"""

    _priv_classes: Dict[bytes, object] = {}

    if ed25519_available: # pragma: no branch
        _priv_classes[b'ed25519'] = ed25519.Ed25519PrivateKey

    if ed448_available: # pragma: no branch
        _priv_classes[b'ed448'] = ed448.Ed448PrivateKey

    @classmethod
    def construct(cls, curve_id: bytes, priv: bytes) -> 'EdDSAPrivateKey':
        """Construct an EdDSA private key"""

        priv_cls = cast('_EdPrivateKey', cls._priv_classes[curve_id])
        priv_key = priv_cls.from_private_bytes(priv)
        pub_key = priv_key.public_key()
        pub = pub_key.public_bytes(Encoding.Raw, PublicFormat.Raw)

        return cls(priv_key, pub, priv)

    @classmethod
    def generate(cls, curve_id: bytes) -> 'EdDSAPrivateKey':
        """Generate a new EdDSA private key"""

        priv_cls = cast('_EdPrivateKey', cls._priv_classes[curve_id])
        priv_key = priv_cls.generate()
        priv = priv_key.private_bytes(Encoding.Raw, PrivateFormat.Raw,
                                      NoEncryption())

        pub_key = priv_key.public_key()
        pub = pub_key.public_bytes(Encoding.Raw, PublicFormat.Raw)

        return cls(priv_key, pub, priv)

    def sign(self, data: bytes, hash_name: str = '') -> bytes:
        """Sign a block of data"""

        # pylint: disable=unused-argument

        priv_key = cast('_EdPrivateKey', self.pyca_key)
        return priv_key.sign(data)


class EdDSAPublicKey(_EdDSAKey):
    """A shim around PyCA for EdDSA public keys"""

    _pub_classes: Dict[bytes, object] = {
        b'ed25519': ed25519.Ed25519PublicKey,
        b'ed448': ed448.Ed448PublicKey
    }

    @classmethod
    def construct(cls, curve_id: bytes, pub: bytes) -> 'EdDSAPublicKey':
        """Construct an EdDSA public key"""

        pub_cls = cast('_EdPublicKey', cls._pub_classes[curve_id])
        pub_key = pub_cls.from_public_bytes(pub)

        return cls(pub_key, pub)

    def verify(self, data: bytes, sig: bytes, hash_name: str = '') -> bool:
        """Verify the signature on a block of data"""

        # pylint: disable=unused-argument

        try:
            pub_key = cast('_EdPublicKey', self.pyca_key)
            pub_key.verify(sig, data)
            return True
        except InvalidSignature:
            return False


class Curve25519DH:
    """Curve25519 Diffie Hellman implementation based on PyCA"""

    def __init__(self) -> None:
        self._priv_key = x25519.X25519PrivateKey.generate()

    def get_public(self) -> bytes:
        """Return the public key to send in the handshake"""

        return self._priv_key.public_key().public_bytes(Encoding.Raw,
                                                        PublicFormat.Raw)

    def get_shared_bytes(self, peer_public: bytes) -> bytes:
        """Return the shared key from the peer's public key as bytes"""

        peer_key = x25519.X25519PublicKey.from_public_bytes(peer_public)
        return self._priv_key.exchange(peer_key)

    def get_shared(self, peer_public: bytes) -> int:
        """Return the shared key from the peer's public key"""

        return int.from_bytes(self.get_shared_bytes(peer_public), 'big')


class Curve448DH:
    """Curve448 Diffie Hellman implementation based on PyCA"""

    def __init__(self) -> None:
        self._priv_key = x448.X448PrivateKey.generate()

    def get_public(self) -> bytes:
        """Return the public key to send in the handshake"""

        return self._priv_key.public_key().public_bytes(Encoding.Raw,
                                                        PublicFormat.Raw)

    def get_shared(self, peer_public: bytes) -> int:
        """Return the shared key from the peer's public key"""

        peer_key = x448.X448PublicKey.from_public_bytes(peer_public)
        shared = self._priv_key.exchange(peer_key)
        return int.from_bytes(shared, 'big')



# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""ECDSA public key encryption handler"""

from typing import Dict, Optional, Tuple, Union, cast

from .asn1 import ASN1DecodeError, BitString, ObjectIdentifier, TaggedDERObject
from .asn1 import der_encode, der_decode
from .crypto import CryptoKey, ECDSAPrivateKey, ECDSAPublicKey
from .crypto import lookup_ec_curve_by_params
from .packet import MPInt, String, SSHPacket
from .public_key import SSHKey, SSHOpenSSHCertificateV01
from .public_key import KeyImportError, KeyExportError
from .public_key import register_public_key_alg, register_certificate_alg
from .public_key import register_x509_certificate_alg


_PrivateKeyArgs = Tuple[bytes, Union[bytes, int], bytes]
_PublicKeyArgs = Tuple[bytes, bytes]


# OID for EC prime fields
PRIME_FIELD = ObjectIdentifier('1.2.840.10045.1.1')

_hash_algs = {b'1.3.132.0.10': 'sha256',
              b'nistp256':     'sha256',
              b'nistp384':     'sha384',
              b'nistp521':     'sha512'}

_alg_oids: Dict[bytes, ObjectIdentifier] = {}
_alg_oid_map: Dict[ObjectIdentifier, bytes] = {}


class _ECKey(SSHKey):
    """Handler for elliptic curve public key encryption"""

    _key: Union[ECDSAPrivateKey, ECDSAPublicKey]

    default_x509_hash = 'sha256'
    pem_name = b'EC'
    pkcs8_oid = ObjectIdentifier('1.2.840.10045.2.1')

    def __init__(self, key: CryptoKey):
        super().__init__(key)

        self.algorithm = b'ecdsa-sha2-' + self._key.curve_id
        self.sig_algorithms = (self.algorithm,)
        self.x509_algorithms = (b'x509v3-' + self.algorithm,)
        self.all_sig_algorithms = set(self.sig_algorithms)

        self._alg_oid = _alg_oids[self._key.curve_id]
        self._hash_alg = _hash_algs[self._key.curve_id]

    def __eq__(self, other: object) -> bool:
        # This isn't protected access - both objects are _ECKey instances
        # pylint: disable=protected-access

        return (isinstance(other, type(self)) and
                self._key.curve_id == other._key.curve_id and
                self._key.x == other._key.x and
                self._key.y == other._key.y and
                self._key.d == other._key.d)

    def __hash__(self) -> int:
        return hash((self._key.curve_id, self._key.x,
                     self._key.y, self._key.d))

    @classmethod
    def _lookup_curve(cls, alg_params: object) -> bytes:
        """Look up an EC curve matching the specified parameters"""

        if isinstance(alg_params, ObjectIdentifier):
            try:
                curve_id = _alg_oid_map[alg_params]
            except KeyError:
                raise KeyImportError('Unknown elliptic curve OID '
                                     f'{alg_params}') from None
        elif (isinstance(alg_params, tuple) and len(alg_params) >= 5 and
              alg_params[0] == 1 and isinstance(alg_params[1], tuple) and
              len(alg_params[1]) == 2 and alg_params[1][0] == PRIME_FIELD and
              isinstance(alg_params[2], tuple) and len(alg_params[2]) >= 2 and
              isinstance(alg_params[3], bytes) and
              isinstance(alg_params[2][0], bytes) and
              isinstance(alg_params[2][1], bytes) and
              isinstance(alg_params[4], int)):
            p = alg_params[1][1]
            a = int.from_bytes(alg_params[2][0], 'big')
            b = int.from_bytes(alg_params[2][1], 'big')
            point = alg_params[3]
            n = alg_params[4]

            try:
                curve_id = lookup_ec_curve_by_params(p, a, b, point, n)
            except ValueError as exc:
                raise KeyImportError(str(exc)) from None
        else:
            raise KeyImportError('Invalid EC curve parameters')

        return curve_id

    @classmethod
    def generate(cls, algorithm: bytes) -> '_ECKey': # type: ignore
        """Generate a new EC private key"""

        # pylint: disable=arguments-differ

        # Strip 'ecdsa-sha2-' prefix of algorithm to get curve_id
        return cls(ECDSAPrivateKey.generate(algorithm[11:]))

    @classmethod
    def make_private(cls, key_params: object) -> SSHKey:
        """Construct an EC private key"""

        curve_id, private_value, public_value = \
            cast(_PrivateKeyArgs, key_params)

        if isinstance(private_value, bytes):
            private_value = int.from_bytes(private_value, 'big')

        return cls(ECDSAPrivateKey.construct(curve_id, public_value,
                                             private_value))

    @classmethod
    def make_public(cls, key_params: object) -> SSHKey:
        """Construct an EC public key"""

        curve_id, public_value = cast(_PublicKeyArgs, key_params)

        return cls(ECDSAPublicKey.construct(curve_id, public_value))

    @classmethod
    def decode_pkcs1_private(cls, key_data: object) -> \
            Optional[_PrivateKeyArgs]:
        """Decode a PKCS#1 format EC private key"""

        if (isinstance(key_data, tuple) and len(key_data) > 2 and
                key_data[0] == 1 and isinstance(key_data[1], bytes) and
                isinstance(key_data[2], TaggedDERObject) and
                key_data[2].tag == 0):
            alg_params = key_data[2].value
            private_key = key_data[1]

            if (len(key_data) > 3 and
                    isinstance(key_data[3], TaggedDERObject) and
                    key_data[3].tag == 1 and
                    isinstance(key_data[3].value, BitString) and
                    key_data[3].value.unused == 0):
                public_key: bytes = key_data[3].value.value
            else:
                public_key = b''

            return cls._lookup_curve(alg_params), private_key, public_key
        else:
            return None

    @classmethod
    def decode_pkcs1_public(cls, key_data: object) -> \
            Optional[_PublicKeyArgs]:
        """Decode a PKCS#1 format EC public key"""

        # pylint: disable=unused-argument

        raise KeyImportError('PKCS#1 not supported for EC public keys')

    @classmethod
    def decode_pkcs8_private(cls, alg_params: object,
                             data: bytes) -> Optional[_PrivateKeyArgs]:
        """Decode a PKCS#8 format EC private key"""

        try:
            key_data = der_decode(data)
        except ASN1DecodeError:
            key_data = None

        if (isinstance(key_data, tuple) and len(key_data) > 1 and
                key_data[0] == 1 and isinstance(key_data[1], bytes)):
            private_key = key_data[1]

            if (len(key_data) > 2 and
                    isinstance(key_data[2], TaggedDERObject) and
                    key_data[2].tag == 1 and
                    isinstance(key_data[2].value, BitString) and
                    key_data[2].value.unused == 0):
                public_key = key_data[2].value.value
            else:
                public_key = b''

            return cls._lookup_curve(alg_params), private_key, public_key
        else:
            return None

    @classmethod
    def decode_pkcs8_public(cls, alg_params: object,
                            data: bytes) -> Optional[_PublicKeyArgs]:
        """Decode a PKCS#8 format EC public key"""

        if isinstance(alg_params, ObjectIdentifier):
            return cls._lookup_curve(alg_params), data
        else:
            return None

    @classmethod
    def decode_ssh_private(cls, packet: SSHPacket) -> _PrivateKeyArgs:
        """Decode an SSH format EC private key"""

        curve_id = packet.get_string()
        public_key = packet.get_string()
        private_key = packet.get_mpint()

        return curve_id, private_key, public_key

    @classmethod
    def decode_ssh_public(cls, packet: SSHPacket) -> _PublicKeyArgs:
        """Decode an SSH format EC public key"""

        curve_id = packet.get_string()
        public_key = packet.get_string()

        return curve_id, public_key

    def encode_public_tagged(self) -> object:
        """Encode an EC public key blob as a tagged bitstring"""

        return TaggedDERObject(1, BitString(self._key.public_value))

    def encode_pkcs1_private(self) -> object:
        """Encode a PKCS#1 format EC private key"""

        if not self._key.private_value:
            raise KeyExportError('Key is not private')

        return (1, self._key.private_value,
                TaggedDERObject(0, self._alg_oid),
                self.encode_public_tagged())

    def encode_pkcs1_public(self) -> object:
        """Encode a PKCS#1 format EC public key"""

        raise KeyExportError('PKCS#1 is not supported for EC public keys')

    def encode_pkcs8_private(self) -> Tuple[object, object]:
        """Encode a PKCS#8 format EC private key"""

        if not self._key.private_value:
            raise KeyExportError('Key is not private')

        return self._alg_oid, der_encode((1, self._key.private_value,
                                          self.encode_public_tagged()))

    def encode_pkcs8_public(self) -> Tuple[object, object]:
        """Encode a PKCS#8 format EC public key"""

        return self._alg_oid, self._key.public_value

    def encode_ssh_private(self) -> bytes:
        """Encode an SSH format EC private key"""

        if not self._key.d:
            raise KeyExportError('Key is not private')

        return b''.join((String(self._key.curve_id),
                         String(self._key.public_value),
                         MPInt(self._key.d)))

    def encode_ssh_public(self) -> bytes:
        """Encode an SSH format EC public key"""

        return b''.join((String(self._key.curve_id),
                         String(self._key.public_value)))

    def encode_agent_cert_private(self) -> bytes:
        """Encode ECDSA certificate private key data for agent"""

        if not self._key.d:
            raise KeyExportError('Key is not private')

        return MPInt(self._key.d)

    def sign_ssh(self, data: bytes, sig_algorithm: bytes) -> bytes:
        """Compute an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        if not self._key.private_value:
            raise ValueError('Private key needed for signing')

        sig = der_decode(self._key.sign(data, self._hash_alg))
        r, s = cast(Tuple[int, int], sig)
        return String(MPInt(r) + MPInt(s))

    def verify_ssh(self, data: bytes, sig_algorithm: bytes,
                   packet: SSHPacket) -> bool:
        """Verify an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        sig = packet.get_string()
        packet.check_end()

        packet = SSHPacket(sig)
        r = packet.get_mpint()
        s = packet.get_mpint()
        packet.check_end()

        return self._key.verify(data, der_encode((r, s)), self._hash_alg)


for _curve_id, _oid_str in ((b'nistp521', '1.3.132.0.35'),
                            (b'nistp384', '1.3.132.0.34'),
                            (b'nistp256', '1.2.840.10045.3.1.7'),
                            (b'1.3.132.0.10', '1.3.132.0.10')):
    _algorithm = b'ecdsa-sha2-' + _curve_id
    _cert_algorithm = _algorithm + b'-cert-v01@openssh.com'
    _x509_algorithm = b'x509v3-' + _algorithm

    _oid = ObjectIdentifier(_oid_str)
    _alg_oids[_curve_id] = _oid
    _alg_oid_map[_oid] = _curve_id

    register_public_key_alg(_algorithm, _ECKey, True, (_algorithm,))
    register_certificate_alg(1, _algorithm, _cert_algorithm,
                             _ECKey, SSHOpenSSHCertificateV01, True)
    register_x509_certificate_alg(_x509_algorithm, True)



# Copyright (c) 2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH TUN/TAP forwarding support"""

import asyncio
import errno
import os
import socket
import struct
import sys
import threading

from typing import Callable, Optional, Tuple, cast

if sys.platform != 'win32': # pragma: no branch
    import fcntl


SSH_TUN_MODE_POINTTOPOINT = 1   # layer 3 IP packets
SSH_TUN_MODE_ETHERNET = 2       # layer 2 Ethenet frames

SSH_TUN_UNIT_ANY = 0x7fffffff   # The server may choose the unit

SSH_TUN_AF_INET = 2             # IPv4
SSH_TUN_AF_INET6 = 24           # IPv6

DARWIN_CTLIOCGINFO = 0xc0644e03
DARWIN_CTLIOCGINFO_FMT = 'I96s'

DARWIN_SIOCGIFFLAGS = 0xc0206911
DARWIN_SIOCSIFFLAGS = 0x80206910

LINUX_TUNSETIFF = 0x400454ca
LINUX_IFF_TUN = 0x1
LINUX_IFF_TAP = 0x2
LINUX_IFF_NO_PI = 0x1000

IFF_FMT = '16sH'
IFF_UP = 0x1


class SSHTunTapTransport(asyncio.Transport):
    """Layer 2/3 tunnel transport"""

    def __init__(self, loop: asyncio.AbstractEventLoop, interface: str):
        super().__init__(extra={'interface': interface})

        self._loop = loop
        self._protocol: Optional[asyncio.Protocol] = None

    def get_protocol(self) -> asyncio.BaseProtocol: # pragma: no cover
        """Get protocol object associated with transport"""

        assert self._protocol is not None

        return self._protocol

    def set_protocol(self, protocol: asyncio.BaseProtocol) -> None:
        """Set protocol associated with transport"""

        self._protocol = cast(asyncio.Protocol, protocol)

    def abort(self) -> None: # pragma: no cover
        """Abort this transport"""

        self.close()

    def is_reading(self) -> bool:
        """Return if the transport is reading data"""

        raise NotImplementedError

    def pause_reading(self) -> None:
        """Pause reading"""

        raise NotImplementedError

    def resume_reading(self) -> None:
        """Resume reading"""

        raise NotImplementedError

    def can_write_eof(self) -> bool: # pragma: no cover
        """This transport doesn't support writing EOF"""

        return False

    def get_write_buffer_size(self) -> int: # pragma: no cover
        """This transport has no output buffer"""

        return 0

    def get_write_buffer_limits(self) -> Tuple[int, int]: # pragma: no cover
        """This transport doesn't support write buffer limits"""

        return 0, 0

    def set_write_buffer_limits(self, high: Optional[int] = None,
                                low: Optional[int] = None) -> None:
        """This transport doesn't support write buffer limits"""

    def write_eof(self) -> None:
        """Ignore writing EOF on this transport"""

    def write(self, data: bytes) -> None:
        """Write a packet"""

        raise NotImplementedError

    def is_closing(self) -> bool: # pragma: no cover
        """Return if the transport is closing"""

        return False

    def close(self) -> None:
        """Close this transport"""

        raise NotImplementedError


class SSHTunTapOSXTransport(SSHTunTapTransport):
    """TunTapOSX transport"""

    def __init__(self, loop: asyncio.AbstractEventLoop, mode: int,
                 unit: Optional[int]):
        prefix = 'tun' if mode == SSH_TUN_MODE_POINTTOPOINT else 'tap'

        if unit is None:
            for i in range(16):
                try:
                    file = open(f'/dev/{prefix}{i}', 'rb+', buffering=0)
                except OSError:
                    pass
                else:
                    unit = i
                    break
            else:
                raise OSError(errno.EBUSY, f'No {prefix} devices available')
        else:
            file = open(f'/dev/{prefix}{unit}', 'rb+', buffering=0)

        interface = f'{prefix}{unit}'
        name = interface.encode()

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        try:
            ifr = struct.pack(IFF_FMT, name, 0)
            ifr = fcntl.ioctl(sock, DARWIN_SIOCGIFFLAGS, ifr)

            _, flags = struct.unpack(IFF_FMT, ifr)
            flags |= IFF_UP

            ifr = struct.pack(IFF_FMT, name, flags)
            fcntl.ioctl(sock, DARWIN_SIOCSIFFLAGS, ifr)
        finally:
            sock.close()

        super().__init__(loop, interface)

        self._file = file
        self._read_thread: Optional[threading.Thread] = None
        os.set_blocking(file.fileno(), True)

    def is_reading(self) -> bool:
        """Return if the transport is reading data"""

        return self._read_thread is not None # pragma: no cover

    def pause_reading(self) -> None:
        """Pause reading"""

        if self._read_thread: # pragma: no branch
            self._read_thread.join()
            self._read_thread = None

    def resume_reading(self) -> None:
        """Resume reading"""

        if not self._read_thread: # pragma: no branch
            self._read_thread = threading.Thread(target=self._read_loop)
            self._read_thread.daemon = True
            self._read_thread.start()

    def _read_loop(self) -> None:
        """Loop reading packets until read is paused or done"""

        assert self._protocol is not None

        while True:
            try:
                data = self._file.read(65536)
            except OSError as exc:
                if exc.errno != errno.EBADF: # pragma: no cover
                    self._loop.call_soon_threadsafe(
                        self._protocol.connection_lost, exc)

                break
            else:
                self._loop.call_soon_threadsafe(
                    self._protocol.data_received, data)

    def write(self, data: bytes) -> None:
        """Write a packet"""

        self._file.write(data)

    def close(self) -> None:
        """Close this transport"""

        self._file.close()
        self.pause_reading()


class SSHDarwinUTunTransport(SSHTunTapTransport):
    """Darwin UTun transport"""

    def __init__(self, loop: asyncio.AbstractEventLoop, unit: Optional[int]):
        sock = socket.socket(socket.PF_SYSTEM, socket.SOCK_DGRAM,
                             socket.SYSPROTO_CONTROL)

        try:
            arg = struct.pack(DARWIN_CTLIOCGINFO_FMT, 0,
                              b'com.apple.net.utun_control')

            ctl_info = fcntl.ioctl(sock, DARWIN_CTLIOCGINFO, arg)
            ctl_id, _ = struct.unpack(DARWIN_CTLIOCGINFO_FMT, ctl_info)

            unit = 0 if unit is None else unit - 15

            sock.setblocking(False)
            sock.connect((ctl_id, unit))

            _, unit = sock.getpeername()
        except OSError:
            sock.close()
            raise

        unit: int

        super().__init__(loop, f'utun{unit-1}')

        self._sock = sock
        self._reading = False

    def is_reading(self) -> bool: # pragma: no cover
        """Return if the transport is reading data"""

        return self._reading

    def pause_reading(self) -> None:
        """Pause reading"""

        self._reading = False
        self._loop.remove_reader(self._sock)

    def resume_reading(self) -> None:
        """Resume reading"""

        self._reading = True
        self._loop.add_reader(self._sock, self._read_ready)

    def _read_ready(self) -> None:
        """Read available packets from the transport"""

        assert self._protocol is not None

        while True:
            try:
                data = self._sock.recv(65540)[4:]
            except (BlockingIOError, InterruptedError):
                break
            except OSError as exc: # pragma: no cover
                self._protocol.connection_lost(exc)
                break
            else:
                self._protocol.data_received(data)

    def write(self, data: bytes) -> None:
        """Write a packet"""

        version = data[0] >> 4
        family = socket.AF_INET if version == 4 else socket.AF_INET6
        data = family.to_bytes(4, 'big') + data

        self._sock.send(data)

    def close(self) -> None:
        """Close this transport"""

        self._sock.close()
        self.pause_reading()


class SSHLinuxTunTapTransport(SSHTunTapTransport):
    """Linux TUN/TAP transport"""

    def __init__(self, loop: asyncio.AbstractEventLoop, mode: int,
                 unit: Optional[int]):
        file = open('/dev/net/tun', 'rb+', buffering=0)

        if mode == SSH_TUN_MODE_POINTTOPOINT:
            flags = LINUX_IFF_TUN | LINUX_IFF_NO_PI
            prefix = 'tun'
        else:
            flags = LINUX_IFF_TAP | LINUX_IFF_NO_PI
            prefix = 'tap'

        name = b'' if unit is None else f'{prefix}{unit}'.encode()

        ifr = struct.pack(IFF_FMT, name, flags)

        try:
            ifr = fcntl.ioctl(file, LINUX_TUNSETIFF, ifr)
        except OSError:
            file.close()
            raise

        name, _ = struct.unpack(IFF_FMT, ifr)
        interface = name.strip(b'\0').decode()

        super().__init__(loop, interface)

        self._file = file
        self._reading = False
        os.set_blocking(file.fileno(), False)

    def is_reading(self) -> bool: # pragma: no cover
        """Return if the transport is reading data"""

        return self._reading

    def pause_reading(self) -> None:
        """Pause reading"""

        self._reading = False

        try:
            self._loop.remove_reader(self._file)
        except OSError: # pragma: no cover
            pass

    def resume_reading(self) -> None:
        """Resume reading"""

        self._reading = True
        self._loop.add_reader(self._file, self._read_ready)

    def _read_ready(self) -> None:
        """Read available packets from the transport"""

        assert self._protocol is not None

        while True:
            try:
                data = self._file.read(65536)
            except OSError as exc: # pragma: no cover
                self._protocol.connection_lost(exc)
                break
            else:
                if data is None:
                    break

                self._protocol.data_received(data)

    def write(self, data: bytes) -> None:
        """Write a packet"""

        self._file.write(data)

    def close(self) -> None:
        """Close this transport"""

        self._file.close()
        self.pause_reading()


def create_tuntap(protocol_factory: Callable[[], asyncio.BaseProtocol],
                  mode: int, unit: Optional[int]) -> \
        Tuple[SSHTunTapTransport, asyncio.BaseProtocol]:
    """Create a local TUN or TAP network interface"""

    loop = asyncio.get_event_loop()
    transport: Optional[SSHTunTapTransport] = None

    if sys.platform == 'darwin':
        if unit is None:
            try:
                transport = SSHTunTapOSXTransport(loop, mode, unit)
            except OSError:
                if mode == SSH_TUN_MODE_POINTTOPOINT:
                    transport = SSHDarwinUTunTransport(loop, unit)
                else:
                    raise
        elif mode == SSH_TUN_MODE_POINTTOPOINT and unit >= 16:
            transport = SSHDarwinUTunTransport(loop, unit)
        else:
            transport = SSHTunTapOSXTransport(loop, mode, unit)
    elif sys.platform == 'linux':
        transport = SSHLinuxTunTapTransport(loop, mode, unit)
    else:
        raise OSError(errno.EPROTONOSUPPORT,
                      f'TunTap not supported on {sys.platform}')

    assert transport is not None

    protocol = protocol_factory()
    protocol.connection_made(transport)

    transport.set_protocol(protocol)
    transport.resume_reading()

    return transport, protocol



# Copyright (c) 2013-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH message authentication handlers"""

from hashlib import md5, sha1, sha224, sha256, sha384, sha512
import hmac
from typing import Dict, Callable, List, Tuple

from .packet import UInt32, UInt64

try:
    from .crypto import umac64, umac128
    _umac_available = True
except ImportError: # pragma: no cover
    _umac_available = False


_MACAlgsArgs = Tuple[bytes, int, int, bool, Callable, Tuple, bool]
_MACHandler = Tuple[Callable, int, Tuple]
_MACParams = Tuple[int, int, bool]


_OPENSSH = b'@openssh.com'
_ETM = b'-etm' + _OPENSSH

_mac_algs: List[bytes] = []
_default_mac_algs: List[bytes] = []
_mac_handler: Dict[bytes, _MACHandler] = {}
_mac_params: Dict[bytes, _MACParams] = {}


class MAC:
    """Parent class for SSH message authentication handlers"""

    def __init__(self, key: bytes, hash_size: int):
        self._key = key
        self._hash_size = hash_size

    def sign(self, seq: int, packet: bytes) -> bytes:
        """Compute a signature for a message"""

        raise NotImplementedError

    def verify(self, seq: int, packet: bytes, sig: bytes) -> bool:
        """Verify the signature of a message"""

        raise NotImplementedError


class _NullMAC(MAC):
    """Null message authentication handler"""

    def sign(self, seq: int, packet: bytes) -> bytes:
        """Compute a signature for a message"""

        return b''

    def verify(self, seq: int, packet: bytes, sig: bytes) -> bool:
        """Verify the signature of a message"""

        return sig == b''


class _HMAC(MAC):
    """HMAC-based message authentication handler"""

    def __init__(self, key: bytes, hash_size: int, hash_alg: Callable):
        super().__init__(key, hash_size)
        self._hash_alg = hash_alg

    def sign(self, seq: int, packet: bytes) -> bytes:
        """Compute a signature for a message"""

        data = UInt32(seq) + packet
        sig = hmac.new(self._key, data, self._hash_alg).digest()
        return sig[:self._hash_size]

    def verify(self, seq: int, packet: bytes, sig: bytes) -> bool:
        """Verify the signature of a message"""

        return hmac.compare_digest(self.sign(seq, packet), sig)


class _UMAC(MAC):
    """UMAC-based message authentication handler"""

    def __init__(self, key: bytes, hash_size: int, umac_alg: Callable):
        super().__init__(key, hash_size)
        self._umac_alg = umac_alg

    def sign(self, seq: int, packet: bytes) -> bytes:
        """Compute a signature for a message"""

        return self._umac_alg(self._key, packet, UInt64(seq)).digest()

    def verify(self, seq: int, packet: bytes, sig: bytes) -> bool:
        """Verify the signature of a message"""

        return hmac.compare_digest(self.sign(seq, packet), sig)


def register_mac_alg(mac_alg: bytes, key_size: int, hash_size: int,
                     etm: bool, handler: Callable, args: Tuple,
                     default: bool) -> None:
    """Register a MAC algorithm"""

    if mac_alg:
        _mac_algs.append(mac_alg)

        if default:
            _default_mac_algs.append(mac_alg)

    _mac_handler[mac_alg] = (handler, hash_size, args)
    _mac_params[mac_alg] = (key_size, hash_size, etm)


def get_mac_algs() -> List[bytes]:
    """Return supported MAC algorithms"""

    return _mac_algs


def get_default_mac_algs() -> List[bytes]:
    """Return default MAC algorithms"""

    return _default_mac_algs


def get_mac_params(mac_alg: bytes) -> _MACParams:
    """Get parameters of a MAC algorithm

       This function returns the key and hash sizes of a MAC algorithm and
       whether or not to compute the MAC before or after encryption.

    """

    return _mac_params[mac_alg]


def get_mac(mac_alg: bytes, key: bytes) -> MAC:
    """Return a MAC handler

       This function returns a MAC object initialized with the specified
       key that can be used for data signing and verification.

    """

    handler, hash_size, args = _mac_handler[mac_alg]
    return handler(key, hash_size, *args)


_mac_algs_list: Tuple[_MACAlgsArgs, ...] = (
    (b'',                         0,  0, False, _NullMAC, (),         True),
)

if _umac_available: # pragma: no branch
    _mac_algs_list += (
        (b'umac-64' + _ETM,      16,  8, True,  _UMAC,    (umac64,),  True),
        (b'umac-128' + _ETM,     16, 16, True,  _UMAC,    (umac128,), True))

_mac_algs_list += (
    (b'hmac-sha2-256' + _ETM,    32, 32, True,  _HMAC,    (sha256,),  True),
    (b'hmac-sha2-512' + _ETM,    64, 64, True,  _HMAC,    (sha512,),  True),
    (b'hmac-sha1' + _ETM,        20, 20, True,  _HMAC,    (sha1,),    True),
    (b'hmac-md5' + _ETM,         16, 16, True,  _HMAC,    (md5,),     False),
    (b'hmac-sha2-256-96' + _ETM, 32, 12, True,  _HMAC,    (sha256,),  False),
    (b'hmac-sha2-512-96' + _ETM, 64, 12, True,  _HMAC,    (sha512,),  False),
    (b'hmac-sha1-96' + _ETM,     20, 12, True,  _HMAC,    (sha1,),    False),
    (b'hmac-md5-96' + _ETM,      16, 12, True,  _HMAC,    (md5,),     False))

if _umac_available: # pragma: no branch
    _mac_algs_list += (
        (b'umac-64' + _OPENSSH,  16,  8, False, _UMAC,    (umac64,),  True),
        (b'umac-128' + _OPENSSH, 16, 16, False, _UMAC,    (umac128,), True))

_mac_algs_list += (
    (b'hmac-sha2-256',           32, 32, False, _HMAC,    (sha256,),  True),
    (b'hmac-sha2-512',           64, 64, False, _HMAC,    (sha512,),  True),
    (b'hmac-sha1',               20, 20, False, _HMAC,    (sha1,),    True),
    (b'hmac-sha256-2@ssh.com',   32, 32, False, _HMAC,    (sha256,),  True),
    (b'hmac-sha224@ssh.com',     28, 28, False, _HMAC,    (sha224,),  True),
    (b'hmac-sha256@ssh.com',     16, 32, False, _HMAC,    (sha256,),  True),
    (b'hmac-sha384@ssh.com',     48, 48, False, _HMAC,    (sha384,),  True),
    (b'hmac-sha512@ssh.com',     64, 64, False, _HMAC,    (sha512,),  True),
    (b'hmac-md5',                16, 16, False, _HMAC,    (md5,),     False),
    (b'hmac-sha2-256-96',        32, 12, False, _HMAC,    (sha256,),  False),
    (b'hmac-sha2-512-96',        64, 12, False, _HMAC,    (sha512,),  False),
    (b'hmac-sha1-96',            20, 12, False, _HMAC,    (sha1,),    False),
    (b'hmac-md5-96',             16, 12, False, _HMAC,    (md5,),     False))

for _alg, _key_size, _hash_size, _etm, \
        _mac_alg, _args, _default in _mac_algs_list:
    register_mac_alg(_alg, _key_size, _hash_size, _etm,
                     _mac_alg, _args, _default)



# Copyright (c) 2018-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH keysign client"""

import asyncio
from pathlib import Path
import subprocess
from typing import Iterable, Sequence, Union, cast

from .misc import FilePath
from .packet import Byte, String, UInt32, PacketDecodeError, SSHPacket
from .public_key import SSHKey, SSHKeyPair, SSHCertificate


_KeySignKey = Union[SSHKey, SSHCertificate]
KeySignPath = Union[None, bool, FilePath]


KEYSIGN_VERSION = 2

_DEFAULT_KEYSIGN_DIRS = ('/opt/local/libexec', '/usr/local/libexec',
                         '/usr/libexec', '/usr/libexec/openssh',
                         '/usr/lib/openssh')


class SSHKeySignKeyPair(SSHKeyPair):
    """Surrogate for a key where signing is done via ssh-keysign"""

    def __init__(self, keysign_path: str, sock_fd: int,
                 key_or_cert: _KeySignKey):
        algorithm = key_or_cert.algorithm
        sig_algorithms = key_or_cert.sig_algorithms[:1]
        public_data = key_or_cert.public_data
        comment = key_or_cert.get_comment_bytes()

        super().__init__(algorithm, algorithm, sig_algorithms, sig_algorithms,
                         public_data, comment)

        self._keysign_path = keysign_path
        self._sock_fd = sock_fd

    async def sign_async(self, data: bytes) -> bytes:
        """Use ssh-keysign to sign a block of data with this key"""

        proc = await asyncio.create_subprocess_exec(
            self._keysign_path, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, pass_fds=[self._sock_fd])

        request = String(Byte(KEYSIGN_VERSION) + UInt32(self._sock_fd) +
                         String(data))
        stdout, stderr = await proc.communicate(request)

        if stderr:
            error = stderr.decode().strip()
            raise ValueError(error)

        try:
            packet = SSHPacket(stdout)
            resp = packet.get_string()
            packet.check_end()

            packet = SSHPacket(resp)
            version = packet.get_byte()
            sig = packet.get_string()
            packet.check_end()

            if version != KEYSIGN_VERSION:
                raise ValueError('unexpected version')

            return sig
        except PacketDecodeError:
            raise ValueError('invalid response') from None


def find_keysign(path: KeySignPath) -> str:
    """Return path to ssh-keysign executable"""

    if path is True:
        for keysign_dir in _DEFAULT_KEYSIGN_DIRS:
            path = Path(keysign_dir, 'ssh-keysign')
            if path.exists():
                break
        else:
            raise ValueError('Keysign not found')
    else:
        if not path or not Path(cast(FilePath, path)).exists():
            raise ValueError('Keysign not found')

    return str(path)


def get_keysign_keys(keysign_path: str, sock_fd: int,
                     keys: Iterable[_KeySignKey]) -> \
        Sequence[SSHKeySignKeyPair]:
    """Return keypair objects which invoke ssh-keysign"""

    return [SSHKeySignKeyPair(keysign_path, sock_fd, key) for key in keys]



# Copyright (c) 2017-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""GSSAPI wrapper"""

import sys

from typing import Optional

from .misc import BytesOrStrDict


try:
    # pylint: disable=unused-import

    if sys.platform == 'win32': # pragma: no cover
        from .gss_win32 import GSSBase, GSSClient, GSSServer, GSSError
    else:
        from .gss_unix import GSSBase, GSSClient, GSSServer, GSSError

    gss_available = True
except ImportError: # pragma: no cover
    gss_available = False

    class GSSError(ValueError): # type: ignore
        """Stub class for reporting that GSS is not available"""

        def __init__(self, maj_code: int, min_code: int,
                 token: Optional[bytes] = None):
            super().__init__('GSS not available')

            self.maj_code = maj_code
            self.min_code = min_code
            self.token = token

    class GSSBase: # type: ignore
        """Base class for reporting that GSS is not available"""

    class GSSClient(GSSBase): # type: ignore
        """Stub client class for reporting that GSS is not available"""

        def __init__(self, _host: str, _store: Optional[BytesOrStrDict],
                     _delegate_creds: bool):
            raise GSSError(0, 0)

    class GSSServer(GSSBase): # type: ignore
        """Stub client class for reporting that GSS is not available"""

        def __init__(self, _host: str, _store: Optional[BytesOrStrDict]):
            raise GSSError(0, 0)



# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""An asynchronous SSH2 library for Python"""

from .version import __author__, __author_email__, __url__, __version__

# pylint: disable=wildcard-import

from .constants import *

# pylint: enable=wildcard-import

from .agent import SSHAgentClient, SSHAgentKeyPair, connect_agent

from .auth_keys import SSHAuthorizedKeys
from .auth_keys import import_authorized_keys, read_authorized_keys

from .channel import SSHClientChannel, SSHServerChannel
from .channel import SSHTCPChannel, SSHUNIXChannel, SSHTunTapChannel

from .client import SSHClient

from .config import ConfigParseError

from .forward import SSHForwarder

from .connection import SSHAcceptor, SSHClientConnection, SSHServerConnection
from .connection import SSHClientConnectionOptions, SSHServerConnectionOptions
from .connection import SSHAcceptHandler
from .connection import create_connection, create_server, connect, listen
from .connection import connect_reverse, listen_reverse, get_server_host_key
from .connection import get_server_auth_methods, run_client, run_server

from .editor import SSHLineEditorChannel

from .known_hosts import SSHKnownHosts
from .known_hosts import import_known_hosts, read_known_hosts
from .known_hosts import match_known_hosts

from .listener import SSHListener

from .logging import logger, set_log_level, set_sftp_log_level, set_debug_level

from .misc import BytesOrStr
from .misc import Error, DisconnectError, ChannelOpenError, ChannelListenError
from .misc import ConnectionLost, CompressionError, HostKeyNotVerifiable
from .misc import KeyExchangeFailed, IllegalUserName, MACError
from .misc import PermissionDenied, ProtocolError, ProtocolNotSupported
from .misc import ServiceNotAvailable, PasswordChangeRequired
from .misc import BreakReceived, SignalReceived, TerminalSizeChanged

from .pbe import KeyEncryptionError

from .pkcs11 import load_pkcs11_keys

from .process import SSHServerProcessFactory
from .process import SSHClientProcess, SSHServerProcess
from .process import SSHCompletedProcess, ProcessError
from .process import TimeoutError # pylint: disable=redefined-builtin
from .process import DEVNULL, PIPE, STDOUT

from .public_key import SSHKey, SSHKeyPair, SSHCertificate
from .public_key import KeyGenerationError, KeyImportError, KeyExportError
from .public_key import generate_private_key, import_private_key
from .public_key import import_public_key, import_certificate
from .public_key import read_private_key, read_public_key, read_certificate
from .public_key import read_private_key_list, read_public_key_list
from .public_key import read_certificate_list
from .public_key import load_keypairs, load_public_keys, load_certificates
from .public_key import load_resident_keys

from .rsa import set_default_skip_rsa_key_validation

from .scp import scp

from .session import DataType, SSHClientSession, SSHServerSession
from .session import SSHTCPSession, SSHUNIXSession, SSHTunTapSession

from .server import SSHServer

from .sftp import SFTPClient, SFTPClientFile, SFTPServer, SFTPError
from .sftp import SFTPEOFError, SFTPNoSuchFile, SFTPPermissionDenied
from .sftp import SFTPFailure, SFTPBadMessage, SFTPNoConnection
from .sftp import SFTPInvalidHandle, SFTPNoSuchPath, SFTPFileAlreadyExists
from .sftp import SFTPWriteProtect, SFTPNoMedia, SFTPNoSpaceOnFilesystem
from .sftp import SFTPQuotaExceeded, SFTPUnknownPrincipal, SFTPLockConflict
from .sftp import SFTPDirNotEmpty, SFTPNotADirectory, SFTPInvalidFilename
from .sftp import SFTPLinkLoop, SFTPCannotDelete, SFTPInvalidParameter
from .sftp import SFTPFileIsADirectory, SFTPByteRangeLockConflict
from .sftp import SFTPByteRangeLockRefused, SFTPDeletePending
from .sftp import SFTPFileCorrupt, SFTPOwnerInvalid, SFTPGroupInvalid
from .sftp import SFTPNoMatchingByteRangeLock
from .sftp import SFTPConnectionLost, SFTPOpUnsupported
from .sftp import SFTPAttrs, SFTPVFSAttrs, SFTPName, SFTPLimits
from .sftp import SEEK_SET, SEEK_CUR, SEEK_END

from .stream import SSHSocketSessionFactory, SSHServerSessionFactory
from .stream import SFTPServerFactory, SSHReader, SSHWriter

from .subprocess import SSHSubprocessReadPipe, SSHSubprocessWritePipe
from .subprocess import SSHSubprocessProtocol, SSHSubprocessTransport

# Import these explicitly to trigger register calls in them
from . import sk_eddsa, sk_ecdsa, eddsa, ecdsa, rsa, dsa, kex_dh, kex_rsa

__all__ = [
    '__author__', '__author_email__', '__url__', '__version__',
    'BreakReceived', 'BytesOrStr', 'ChannelListenError',
    'ChannelOpenError', 'CompressionError', 'ConfigParseError',
    'ConnectionLost', 'DEVNULL', 'DataType', 'DisconnectError', 'Error',
    'HostKeyNotVerifiable', 'IllegalUserName', 'KeyEncryptionError',
    'KeyExchangeFailed', 'KeyExportError', 'KeyGenerationError',
    'KeyImportError', 'MACError', 'PIPE', 'PasswordChangeRequired',
    'PermissionDenied', 'ProcessError', 'ProtocolError',
    'ProtocolNotSupported', 'SEEK_CUR', 'SEEK_END', 'SEEK_SET',
    'SFTPAttrs', 'SFTPBadMessage', 'SFTPByteRangeLockConflict',
    'SFTPByteRangeLockRefused', 'SFTPCannotDelete', 'SFTPClient',
    'SFTPClientFile', 'SFTPConnectionLost', 'SFTPDeletePending',
    'SFTPDirNotEmpty', 'SFTPEOFError', 'SFTPError', 'SFTPFailure',
    'SFTPFileAlreadyExists', 'SFTPFileCorrupt', 'SFTPFileIsADirectory',
    'SFTPGroupInvalid', 'SFTPInvalidFilename', 'SFTPInvalidHandle',
    'SFTPInvalidParameter', 'SFTPLimits', 'SFTPLinkLoop', 'SFTPLockConflict',
    'SFTPName', 'SFTPNoConnection', 'SFTPNoMatchingByteRangeLock',
    'SFTPNoMedia', 'SFTPNoSpaceOnFilesystem', 'SFTPNoSuchFile',
    'SFTPNoSuchPath', 'SFTPNotADirectory', 'SFTPOpUnsupported',
    'SFTPOwnerInvalid', 'SFTPPermissionDenied', 'SFTPQuotaExceeded',
    'SFTPServer', 'SFTPServerFactory', 'SFTPUnknownPrincipal', 'SFTPVFSAttrs',
    'SFTPWriteProtect', 'SSHAcceptHandler', 'SSHAcceptor', 'SSHAgentClient',
    'SSHAgentKeyPair', 'SSHAuthorizedKeys', 'SSHCertificate', 'SSHClient',
    'SSHClientChannel', 'SSHClientConnection', 'SSHClientConnectionOptions',
    'SSHClientProcess', 'SSHClientSession', 'SSHCompletedProcess',
    'SSHForwarder', 'SSHKey', 'SSHKeyPair', 'SSHKnownHosts',
    'SSHLineEditorChannel', 'SSHListener', 'SSHReader', 'SSHServer',
    'SSHServerChannel', 'SSHServerConnection',
    'SSHServerConnectionOptions', 'SSHServerProcess',
    'SSHServerProcessFactory', 'SSHServerSession',
    'SSHServerSessionFactory', 'SSHSocketSessionFactory',
    'SSHSubprocessProtocol', 'SSHSubprocessReadPipe',
    'SSHSubprocessTransport', 'SSHSubprocessWritePipe', 'SSHTCPChannel',
    'SSHTCPSession', 'SSHTunTapChannel', 'SSHTunTapSession',
    'SSHUNIXChannel', 'SSHUNIXSession', 'SSHWriter',
    'STDOUT', 'ServiceNotAvailable', 'SignalReceived', 'TerminalSizeChanged',
    'TimeoutError', 'connect', 'connect_agent', 'connect_reverse',
    'create_connection', 'create_server', 'generate_private_key',
    'get_server_auth_methods', 'get_server_host_key',
    'import_authorized_keys', 'import_certificate', 'import_known_hosts',
    'import_private_key', 'import_public_key', 'listen', 'listen_reverse',
    'load_certificates', 'load_keypairs', 'load_pkcs11_keys',
    'load_public_keys', 'load_resident_keys', 'logger', 'match_known_hosts',
    'read_authorized_keys', 'read_certificate', 'read_certificate_list',
    'read_known_hosts', 'read_private_key', 'read_private_key_list',
    'read_public_key', 'read_public_key_list', 'run_client', 'run_server',
    'scp', 'set_debug_level', 'set_default_skip_rsa_key_validation',
    'set_log_level', 'set_sftp_log_level'
]



# Copyright (c) 2016-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH agent support code for UNIX"""

import asyncio
import errno
from typing import TYPE_CHECKING, Tuple


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .agent import AgentReader, AgentWriter


async def open_agent(agent_path: str) -> Tuple['AgentReader', 'AgentWriter']:
    """Open a connection to ssh-agent"""

    if not agent_path:
        raise OSError(errno.ENOENT, 'Agent not found')

    return await asyncio.open_unix_connection(agent_path)



# Copyright (c) 2014-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""A shim for accessing cryptographic primitives needed by asyncssh"""

from .cipher import BasicCipher, GCMCipher, register_cipher, get_cipher_params

from .dsa import DSAPrivateKey, DSAPublicKey

from .dh import DH

from .ec import ECDSAPrivateKey, ECDSAPublicKey, ECDH

from .ed import ed25519_available, ed448_available
from .ed import curve25519_available, curve448_available
from .ed import EdDSAPrivateKey, EdDSAPublicKey, Curve25519DH, Curve448DH

from .ec_params import lookup_ec_curve_by_params

from .kdf import pbkdf2_hmac

from .misc import CryptoKey, PyCAKey

from .rsa import RSAPrivateKey, RSAPublicKey

from .pq import mlkem_available, sntrup_available, PQDH

# Import chacha20-poly1305 cipher if available
from .chacha import ChachaCipher, chacha_available

# Import umac cryptographic hash if available
try:
    from .umac import umac32, umac64, umac96, umac128
except (ImportError, AttributeError, OSError): # pragma: no cover
    pass

# Import X.509 certificate support if available
try:
    from .x509 import X509Certificate, X509Name, X509NamePattern
    from .x509 import generate_x509_certificate, import_x509_certificate
except (ImportError, AttributeError): # pragma: no cover
    pass

__all__ = [
    'BasicCipher', 'ChachaCipher', 'CryptoKey', 'Curve25519DH', 'Curve448DH',
    'DH', 'DSAPrivateKey', 'DSAPublicKey', 'ECDH', 'ECDSAPrivateKey',
    'ECDSAPublicKey', 'EdDSAPrivateKey', 'EdDSAPublicKey', 'GCMCipher', 'PQDH',
    'PyCAKey', 'RSAPrivateKey', 'RSAPublicKey', 'X509Certificate',
    'X509Name', 'X509NamePattern', 'chacha_available', 'curve25519_available',
    'curve448_available', 'ed25519_available', 'ed448_available',
    'generate_x509_certificate', 'get_cipher_params',
    'import_x509_certificate', 'lookup_ec_curve_by_params', 'mlkem_available',
    'pbkdf2_hmac', 'register_cipher', 'sntrup_available', 'umac32', 'umac64',
    'umac96', 'umac128'
]



# Copyright (c) 2020-2023 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""PKCS#11 smart card handler"""

from types import TracebackType
from typing import Dict, List, Optional, Sequence, Tuple, Type, Union, cast

try:
    import pkcs11
    from pkcs11 import Attribute, KeyType, Mechanism, ObjectClass
    from pkcs11 import PrivateKey, Token
    from pkcs11.util.rsa import encode_rsa_public_key
    from pkcs11.util.ec import encode_ec_public_key
    pkcs11_available = True
except (ImportError, ModuleNotFoundError): # pragma: no cover
    pkcs11_available = False

from .misc import BytesOrStr
from .packet import MPInt, String
from .public_key import SSHCertificate, SSHKey, SSHKeyPair
from .public_key import import_certificate_chain, import_public_key


_AttrDict = Dict['Attribute', Union[bool, bytes, str, 'ObjectClass']]
_TokenID = Tuple[str, bytes]
_SessionMap = Dict[_TokenID, 'SSHPKCS11Session']


if pkcs11_available:
    encoders = {KeyType.RSA: encode_rsa_public_key,
                KeyType.EC:  encode_ec_public_key}

    mechanisms = {b'ssh-rsa':                Mechanism.SHA1_RSA_PKCS,
                  b'rsa-sha2-256':           Mechanism.SHA256_RSA_PKCS,
                  b'rsa-sha2-512':           Mechanism.SHA512_RSA_PKCS,
                  b'ssh-rsa-sha224@ssh.com': Mechanism.SHA224_RSA_PKCS,
                  b'ssh-rsa-sha256@ssh.com': Mechanism.SHA256_RSA_PKCS,
                  b'ssh-rsa-sha384@ssh.com': Mechanism.SHA384_RSA_PKCS,
                  b'ssh-rsa-sha512@ssh.com': Mechanism.SHA512_RSA_PKCS,
                  b'rsa1024-sha1':           Mechanism.SHA1_RSA_PKCS,
                  b'rsa2048-sha256':         Mechanism.SHA256_RSA_PKCS,
                  b'ecdsa-sha2-nistp256':    Mechanism.ECDSA_SHA256,
                  b'ecdsa-sha2-nistp384':    Mechanism.ECDSA_SHA384,
                  b'ecdsa-sha2-nistp521':    Mechanism.ECDSA_SHA512}


    class SSHPKCS11KeyPair(SSHKeyPair):
        """Surrogate for a key accessed via a PKCS#11 provider"""

        _key_type = 'pkcs11'

        def __init__(self, session: 'SSHPKCS11Session', privkey: PrivateKey,
                     pubkey: SSHKey, cert: Optional[SSHCertificate] = None):
            super().__init__(pubkey.algorithm, pubkey.algorithm,
                             pubkey.sig_algorithms, pubkey.sig_algorithms,
                             pubkey.public_data, privkey.label, cert,
                             use_executor=True)

            self._session = session
            self._privkey = privkey

        def __del__(self) -> None:
            self._session.close()

        def sign(self, data: bytes) -> bytes:
            """Sign a block of data with this private key"""

            sig_algorithm = self.sig_algorithm

            if sig_algorithm.startswith(b'x509v3-'):
                sig_algorithm = sig_algorithm[7:]

            sig = self._privkey.sign(data, mechanism=mechanisms[sig_algorithm])

            if self._privkey.key_type == KeyType.EC:
                length = len(sig) // 2
                r = int.from_bytes(sig[:length], 'big')
                s = int.from_bytes(sig[length:], 'big')
                sig = MPInt(r) + MPInt(s)

            return String(sig_algorithm) + String(sig)


    class SSHPKCS11Session:
        """Work around PKCS#11 sessions not supporting simultaneous opens"""

        _sessions: _SessionMap = {}

        def __init__(self, token_id: _TokenID, token: Token,
                     pin: Optional[str]):
            self._token_id = token_id
            self._session = token.open(user_pin=pin)
            self._refcount = 0

        def __enter__(self) -> 'SSHPKCS11Session':
            """Allow SSHPKCS11Session to be used as a context manager"""

            return self

        def __exit__(self, _exc_type: Type[BaseException],
                     _exc_value: BaseException,
                     _traceback: TracebackType) -> None:
            """Drop one reference to the session when exiting"""

            self.close()

        @classmethod
        def open(cls, token: Token, pin: Optional[str]) -> 'SSHPKCS11Session':
            """Open a new session, or return an already-open one"""

            token_id = (token.manufacturer_id, token.serial)

            try:
                session = cls._sessions[token_id]
            except KeyError:
                session = cls(token_id, token, pin)
                cls._sessions[token_id] = session

            session._refcount += 1
            return session

        def close(self) -> None:
            """Drop one reference to an open session"""

            self._refcount -= 1

            if self._refcount == 0:
                self._session.close()
                del self._sessions[self._token_id]

        def get_keys(self, load_certs: bool, key_label: Optional[str],
                     key_id: Optional[BytesOrStr]) -> \
                Sequence[SSHPKCS11KeyPair]:
            """Return the private keys found on this token"""

            if isinstance(key_id, str):
                key_id = bytes.fromhex(key_id)

            key_attrs: _AttrDict = {Attribute.CLASS: ObjectClass.PRIVATE_KEY,
                                    Attribute.SIGN: True}

            if key_label is not None:
                key_attrs[Attribute.LABEL] = key_label

            if key_id is not None:
                key_attrs[Attribute.OBJECT_ID] = key_id

            cert_attrs: _AttrDict = {Attribute.CLASS: ObjectClass.CERTIFICATE}

            if load_certs:
                certs = [import_certificate_chain(
                         cast(bytes, cert[Attribute.VALUE]))
                         for cert in self._session.get_objects(cert_attrs)]

                certdict = {cert.key.public_data: cert for cert in certs
                            if cert and 'Attest' not in str(cert.subject)}
            else:
                certdict = {}

            keys = []

            for key in self._session.get_objects(key_attrs):
                privkey = cast(PrivateKey, key)
                encoder = encoders.get(privkey.key_type)

                if encoder:
                    pubkey = import_public_key(encoder(privkey))

                    cert = certdict.get(pubkey.public_data)

                    if cert:
                        keys.append(SSHPKCS11KeyPair(self, privkey,
                                                     pubkey, cert))

                    keys.append(SSHPKCS11KeyPair(self, privkey, pubkey))

            self._refcount += len(keys)

            return keys


    def load_pkcs11_keys(provider: str, pin: Optional[str] = None, *,
                         load_certs: bool = True,
                         token_label: Optional[str] = None,
                         token_serial: Optional[BytesOrStr] = None,
                         key_label: Optional[str] = None,
                         key_id: Optional[BytesOrStr] = None) -> \
            Sequence[SSHPKCS11KeyPair]:
        """Load PIV keys and X.509 certificates from a PKCS#11 token

           This function loads a list of SSH keypairs with optional X.509
           cerificates from attached PKCS#11 security tokens. The PKCS#11
           provider must be specified, along with a user PIN if the
           tokens are set to require one.

           By default, this function loads both private key handles
           and the X.509 certificates associated with them, allowing for
           X.509 certificate based auth to SSH servers that support it.
           To disable loading of these certificates and perform only
           key-based authentication, load_certs may be set to `False`.

           If token_label and/or token_serial are specified, only tokens
           matching those values will be accessed.

           If key_label and/or key_id are specified, only keys matching
           those values will be loaded. Key IDs can be specified as
           either raw bytes or a string containing hex digits.

               .. note:: If you have an active asyncio event loop at
                         the time you call this function, you may want
                         to consider running it via a call to
                         :meth:`asyncio.AbstractEventLoop.run_in_executor`.
                         While retrieving the keys generally takes only a
                         fraction of a second, calling this function
                         directly could block asyncio event processing
                         until it completes.

           :param provider:
               The path to the PKCS#11 provider's shared library.
           :param pin: (optional)
               The PIN to use when accessing tokens, if needed.
           :param load_certs: (optional)
               Whether or not to load X.509 certificates from the
               security tokens.
           :param token_label: (optional)
               A token label to match against. If set, only security
               tokens with this label will be accessed.
           :param token_serial: (optional)
               A token serial number to match against. If set, only
               security tokens with this serial number  will be accessed.
           :param key_label: (optional)
               A key label to match against. If set, only keys with this
               label will be loaded.
           :param key_id: (optional)
               A key ID to match against. If set, only keys with this ID
               will be loaded.
           :type provider: `str`
           :type pin: `str`
           :type load_certs: `bool`
           :type token_label: `str`
           :type token_serial: `bytes` or `str`
           :type key_label: `str`
           :type key_id: `bytes` or `str`

           :returns: list of class:`SSHKeyPair`

        """

        lib = pkcs11.lib(provider)

        keys: List[SSHPKCS11KeyPair] = []

        if isinstance(token_serial, str):
            token_serial = token_serial.encode('utf-8')

        for token in lib.get_tokens(token_label=token_label,
                                    token_serial=token_serial):
            with SSHPKCS11Session.open(token, pin) as session:
                keys.extend(session.get_keys(load_certs, key_label, key_id))

        return keys
else: # pragma: no cover
    def load_pkcs11_keys(provider: str, pin: Optional[str] = None, *,
                         load_certs: bool = True,
                         token_label: Optional[str] = None,
                         token_serial: Optional[BytesOrStr] = None,
                         key_label: Optional[str] = None,
                         key_id: Optional[BytesOrStr] = None) -> \
            Sequence['SSHPKCS11KeyPair']:
        """Report that PKCS#11 support is not available"""

        raise ValueError('PKCS#11 support not available') from None
